@@ -31,6 +31,87 @@ namespace mathutils {
  */
 template <typename T> inline int sign(T x) { return (x > T{0}) - (x < T{0}); }
 
+/**
+ * @brief Convert Cartesian coordinates (x, y, z) to spherical coordinates (r,
+ * theta, phi).
+ *
+ * @param xyz The input Cartesian coordinates [...(x,y,z)...] (N, 3)
+ * @return The spherical coordinates [...(r, theta, phi)...] (N, 3)
+ */
+inline Eigen::MatrixXd rthetaphi_from_xyz(const Eigen::MatrixXd &xyz) {
+  if (xyz.cols() != 3) {
+    throw std::invalid_argument("xyz must have 3 columns");
+  }
+  int num_points = xyz.rows();
+  Eigen::MatrixXd rthetaphi(num_points, 3);
+  rthetaphi.setZero();
+  for (int i = 0; i < num_points; ++i) {
+    double x = xyz(i, 0);
+    double y = xyz(i, 1);
+    double z = xyz(i, 2);
+    double r = std::sqrt(x * x + y * y + z * z);
+    double rho = std::sqrt(x * x + y * y);
+    double theta = std::atan2(rho, z);
+    double phi = std::atan2(y, x);
+    rthetaphi(i, 0) = r;
+    rthetaphi(i, 1) = theta;
+    rthetaphi(i, 2) = phi;
+  }
+  return rthetaphi;
+}
+
+/**
+ * @brief Convert spherical coordinates (r, theta, phi) to Cartesian coordinates
+ * (x, y, z).
+ *
+ * @param rthetaphi The input spherical coordinates [...(r, theta, phi)...] (N,
+ * 3)
+ * @return The Cartesian coordinates [...(x, y, z)...] (N, 3)
+ */
+inline Eigen::MatrixXd xyz_from_rthetaphi(const Eigen::MatrixXd &rthetaphi) {
+  if (rthetaphi.cols() != 3) {
+    throw std::invalid_argument("rthetaphi must have 3 columns");
+  }
+  int num_points = rthetaphi.rows();
+  Eigen::MatrixXd xyz(num_points, 3);
+  xyz.setZero();
+  for (int i = 0; i < num_points; ++i) {
+    double r = rthetaphi(i, 0);
+    double theta = rthetaphi(i, 1);
+    double phi = rthetaphi(i, 2);
+    xyz(i, 0) = r * std::sin(theta) * std::cos(phi);
+    xyz(i, 1) = r * std::sin(theta) * std::sin(phi);
+    xyz(i, 2) = r * std::cos(theta);
+  }
+  return xyz;
+}
+
+/**
+ * @brief Convert Cartesian coordinates (x, y, z) to unit sphere coordinates
+ * (theta, phi) by projecting onto the unit sphere.
+ *
+ * @param xyz The input Cartesian coordinates [...(x,y,z)...] (N, 3)
+ * @return The spherical coordinates [...(theta, phi)...] (N, 2)
+ */
+inline Eigen::MatrixXd thetaphi_from_xyz(const Eigen::MatrixXd &xyz) {
+  if (xyz.cols() != 3) {
+    throw std::invalid_argument("xyz must have 3 columns");
+  }
+  int num_points = xyz.rows();
+  Eigen::MatrixXd thetaphi(num_points, 2);
+  thetaphi.setZero();
+  for (int i = 0; i < num_points; ++i) {
+    double x = xyz(i, 0);
+    double y = xyz(i, 1);
+    double z = xyz(i, 2);
+    double rho = std::sqrt(x * x + y * y);
+    double theta = std::atan2(rho, z);
+    double phi = std::atan2(y, x);
+    thetaphi(i, 0) = theta;
+    thetaphi(i, 1) = phi;
+  }
+  return thetaphi;
+}
 } // namespace mathutils
 /////////////////////////////////////
 /////////////////////////////////////
@@ -145,6 +226,14 @@ inline std::pair<double, int> ReLogRe_ImLogRe_over_pi_of_x_to_pow(const T &x,
 // Spherical harmonics //
 /////////////////////////
 
+/**
+ * @brief Compute the index of (l,m) in [(0,0), (1,-1), (1,0), (1,1), (2,-2),
+ * ...]
+ *
+ * @param l The degree (non-negative).
+ * @param m The order (must be in the range [-l, l]).
+ * @return int The spherical harmonic index.
+ */
 inline int spherical_harmonic_index_n_LM(int l, int m) {
   if (m < -l || m > l) {
     throw std::out_of_range("m must be in the range [-l, l]");
@@ -155,6 +244,14 @@ inline int spherical_harmonic_index_n_LM(int l, int m) {
   return l * (l + 1) + m;
 }
 
+/**
+ * @brief Compute the n-th element of the sequence [(0,0), (1,-1), (1,0), (1,1),
+ * (2,-2),
+ * ...]
+ *
+ * @param n The linear index (must be non-negative).
+ * @return std::array<int, 2> The (l,m) pair.
+ */
 inline std::array<int, 2> spherical_harmonic_index_lm_N(int n) {
   if (n < 0) {
     throw std::out_of_range("n must be non-negative");
@@ -171,7 +268,7 @@ inline std::array<int, 2> spherical_harmonic_index_lm_N(int n) {
  * @brief Compute the associated Legendre polynomial with spherical harmonics
  * normalization \f$P_{m m}(\theta) = e^{-im\phi}Y_{m m}(\theta, \phi)\f$
  * for \f$0\leq\theta\leq\pi/2\f$ and \f$m\geq 0\f$. Used to initialize the
- * recursion.
+ * recursion. Does not check the validity of \f$m\f$ or \f$\theta\f$.
  * @param m Degree/order of the polynomial (\f$m=0,1,\ldots\f$).
  * @param theta Polar angle in radians (\f$0\leq\theta\leq\pi/2\f$).
  * @return double Value of the associated normalized Legendre polynomial.
@@ -195,24 +292,16 @@ inline double reduced_spherical_Pmm(int m, double theta) {
                             0.5 * std::lgamma(2 * m + 2) - std::lgamma(m + 1));
 }
 
-// inline std::pair<double, double>
-// init_Plm_recursion_three_term_upward_ell(int m, double theta) {
-//   double P_m_m =
-//       (theta == 0.0)
-//           ? (m == 0) / std::sqrt(4 * M_PI)
-//           : (1 - 2 * (m & 1)) *
-//                 std::exp(m * std::log(std::sin(theta)) -
-//                          0.5 * std::log(4 * M_PI) - m * std::log(2) +
-//                          0.5 * std::lgamma(2 * m + 2) - std::lgamma(m + 1));
-//   double P_m_plus_one_m = std::sqrt(2 * m + 3) * std::cos(theta) * P_m_m;
-//   return {P_m_m, P_m_plus_one_m};
-// }
-
 /**
- * @brief Generate the associated Legendre polynomial with spherical harmonics
- * normalization \f$P_{\ell m}(\theta) = e^{-im\phi}Y_{\ell m}(\theta,
- * \phi)\f$ for \f$\ell = \ell_{start}, \ell_{start} + 1, \ldots, \ell_{end}\f$
- * and \f$m=\ell_{start}\f$.
+ * @brief Generate a sequence of associated Legendre polynomials with spherical
+ * harmonics normalization \f$P_{\ell m}(\theta) = e^{-im\phi}Y_{\ell m}(\theta,
+ * \phi)\f$ for fixed order \f$m=\ell_{start}\f$ and degree \f$\ell =
+ * \ell_{start}, \ell_{start} + 1, \ldots, \ell_{end}\f$. Does not check the
+ * validity of \f$\theta\f$.
+ *
+ * @param l_start Starting value \f$\ell\f$. Equivalent
+ * @param l_end Final value of \f$\ell\f$.
+ * @param theta Polar angle in radians (\f$0\leq\theta\leq\pi/2\f$).
  */
 mathutils::SimpleGenerator<double>
 generate_reduced_spherical_Plm_recursion_three_term_upward_ell(int l_start,
@@ -249,6 +338,17 @@ generate_reduced_spherical_Plm_recursion_three_term_upward_ell(int l_start,
   }
 }
 
+/**
+ * @brief Generate a sequence of associated Legendre polynomials with spherical
+ * harmonics normalization \f$P_{\ell m}(\theta) = e^{-im\phi}Y_{\ell m}(\theta,
+ * \phi)\f$ for fixed order \f$m=\ell_{start}\f$ and degree \f$\ell =
+ * \ell_{start}, \ell_{start} + 1, \ldots, \ell_{end}\f$. Yields pairs of (l,
+ * P_ell_m). Does not check the validity of \f$\theta\f$.
+ *
+ * @param l_start Starting value \f$\ell\f$. Equivalent
+ * @param l_end Final value of \f$\ell\f$.
+ * @param theta Polar angle in radians (\f$0\leq\theta\leq\pi/2\f$).
+ */
 mathutils::SimpleGenerator<std::pair<int, double>>
 enumerate_reduced_spherical_Plm_recursion_three_term_upward_ell(int l_start,
                                                                 int l_end,
@@ -287,7 +387,8 @@ enumerate_reduced_spherical_Plm_recursion_three_term_upward_ell(int l_start,
 /**
  * @brief Compute the associated Legendre polynomial with spherical harmonics
  * normalization \f$P_{\ell m}(\theta) = e^{-im\phi}Y_{\ell m}(\theta, \phi)\f$
- * for \f$0\leq\theta\leq\pi/2\f$ and \f$m\geq 0\f$.
+ * for \f$0\leq\theta\leq\pi/2\f$ and \f$m\geq 0\f$. Does not check the validity
+ * of inputs.
  *
  * @param l Degree of the polynomial (\f$\ell=0,1,\ldots\f$).
  * @param m Order of the polynomial (\f$m=0,1,\ldots,\ell\f$).
@@ -465,7 +566,9 @@ Eigen::VectorXd real_Ylm(int l, int m,
  * @param thetaphi_coord_P A matrix of shape (num_points, 2) where each row
  * contains the polar angle \f$\theta\f$ and azimuthal angle \f$\phi\f$ in
  * radians.
- * @return A matrix of shape (num_points, l_max * (l_max + 2) + 1)
+ * @return A matrix of shape (num_points, num_modes=l_max * (l_max + 2) + 1).
+ * Mode index is related to (l,m) by
+ * spherical_harmonic_index_n_LM/spherical_harmonic_index_lm_N functions.
  */
 Eigen::MatrixXd compute_all_real_Ylm(int l_max,
                                      const Eigen::MatrixXd &thetaphi_coord_P) {
@@ -522,61 +625,6 @@ Eigen::MatrixXd compute_all_real_Ylm(int l_max,
   return Y;
 }
 
-Eigen::MatrixXcd compute_all_Ylm(int l_max,
-                                 const Eigen::MatrixXd &thetaphi_coord_P) {
-  // check l_max
-  if (l_max < 0) {
-    throw std::out_of_range("l_max must be non-negative");
-  }
-  // check thetaphi_coord_P
-  if (thetaphi_coord_P.cols() != 2) {
-    throw std::invalid_argument("thetaphi_coord_P must have 2 columns");
-  }
-  int num_points = thetaphi_coord_P.rows();
-  if (num_points <= 0) {
-    throw std::invalid_argument("thetaphi_coord_P must have at least one row");
-  }
-  //////////////////////////////////
-  int num_modes = l_max * (l_max + 2) + 1; // = n_LM(l_max, l_max)+1
-  Eigen::MatrixXcd Y(num_points, num_modes);
-  Y.setZero();
-
-  for (int p{0}; p < num_points; ++p) {
-    double theta = thetaphi_coord_P(p, 0);
-    if (theta < 0 || theta > M_PI) {
-      throw std::out_of_range("theta must be in the range [0, pi]");
-    }
-    double phi = thetaphi_coord_P(p, 1);
-    double reduced_theta =
-        std::min(theta, M_PI - theta); // transform theta to [0, pi/2]
-    // m = 0
-    for (auto [l, reduced_P] :
-         enumerate_reduced_spherical_Plm_recursion_three_term_upward_ell(
-             0, l_max, reduced_theta)) {
-      int n = l * (l + 1);
-      Y(p, n) = ((theta > M_PI_2) && (l & 1)) ? -reduced_P : reduced_P;
-    }
-    // m = +/-1, ... , +/-l_max
-    for (int m = 1; m <= l_max; ++m) {
-      for (auto [l, reduced_P] :
-           enumerate_reduced_spherical_Plm_recursion_three_term_upward_ell(
-               m, l_max, reduced_theta)) {
-        int n_plus = l * (l + 1) + m;
-        int n_minus = l * (l + 1) - m;
-        // Y_{l |m|} = P_{l|m|}*exp(i|m|phi)
-        Y(p, n_plus) =
-            ((theta > M_PI_2) && ((l - m) & 1))
-                ? -reduced_P * std::exp(std::complex<double>(0, m * phi))
-                : reduced_P * std::exp(std::complex<double>(0, m * phi));
-        // Y_{l, -|m|} = (-1)^m conj(Y_{l|m|})
-        Y(p, n_minus) =
-            (m & 1) ? -std::conj(Y(p, n_plus)) : std::conj(Y(p, n_plus));
-      }
-    }
-  } // end for p
-  return Y;
-}
-
 /**
  * @brief Compute the spherical harmonics \f$Y_{l m}(\theta, \phi)\f$
  * for \f$0\leq\theta\leq\pi\f$ and \f$-\ell \leq m\geq \ell\f$.
@@ -651,11 +699,83 @@ Eigen::VectorXcd Ylm(int l, int m, const Eigen::MatrixXd &thetaphi_coord_P) {
   return result;
 }
 
-///////////////////////////
-///////////////////////////
-///////////////////////////
-///////////////////////////
+/**
+ * @brief Compute the spherical harmonics \f$Y_{l m}(\theta, \phi)\f$
+ * for \f$0\leq\theta\leq\pi\f$ and \f$-\ell \leq m\geq \ell\f$.
+ *
+ * @param l_max The maximum order of the spherical harmonics (non-negative
+ * integer).
+ * @param thetaphi_coord_P A matrix of shape (num_points, 2) where each row
+ * contains the polar angle \f$\theta\f$ and azimuthal angle \f$\phi\f$ in
+ * radians.
+ * @return A matrix of shape (num_points, num_modes=l_max * (l_max + 2) + 1).
+ * Mode index is related to (l,m) by
+ * spherical_harmonic_index_n_LM/spherical_harmonic_index_lm_N functions.
+ */
+Eigen::MatrixXcd compute_all_Ylm(int l_max,
+                                 const Eigen::MatrixXd &thetaphi_coord_P) {
+  // check l_max
+  if (l_max < 0) {
+    throw std::out_of_range("l_max must be non-negative");
+  }
+  // check thetaphi_coord_P
+  if (thetaphi_coord_P.cols() != 2) {
+    throw std::invalid_argument("thetaphi_coord_P must have 2 columns");
+  }
+  int num_points = thetaphi_coord_P.rows();
+  if (num_points <= 0) {
+    throw std::invalid_argument("thetaphi_coord_P must have at least one row");
+  }
+  //////////////////////////////////
+  int num_modes = l_max * (l_max + 2) + 1; // = n_LM(l_max, l_max)+1
+  Eigen::MatrixXcd Y(num_points, num_modes);
+  Y.setZero();
 
+  for (int p{0}; p < num_points; ++p) {
+    double theta = thetaphi_coord_P(p, 0);
+    if (theta < 0 || theta > M_PI) {
+      throw std::out_of_range("theta must be in the range [0, pi]");
+    }
+    double phi = thetaphi_coord_P(p, 1);
+    double reduced_theta =
+        std::min(theta, M_PI - theta); // transform theta to [0, pi/2]
+    // m = 0
+    for (auto [l, reduced_P] :
+         enumerate_reduced_spherical_Plm_recursion_three_term_upward_ell(
+             0, l_max, reduced_theta)) {
+      int n = l * (l + 1);
+      Y(p, n) = ((theta > M_PI_2) && (l & 1)) ? -reduced_P : reduced_P;
+    }
+    // m = +/-1, ... , +/-l_max
+    for (int m = 1; m <= l_max; ++m) {
+      for (auto [l, reduced_P] :
+           enumerate_reduced_spherical_Plm_recursion_three_term_upward_ell(
+               m, l_max, reduced_theta)) {
+        int n_plus = l * (l + 1) + m;
+        int n_minus = l * (l + 1) - m;
+        // Y_{l |m|} = P_{l|m|}*exp(i|m|phi)
+        Y(p, n_plus) =
+            ((theta > M_PI_2) && ((l - m) & 1))
+                ? -reduced_P * std::exp(std::complex<double>(0, m * phi))
+                : reduced_P * std::exp(std::complex<double>(0, m * phi));
+        // Y_{l, -|m|} = (-1)^m conj(Y_{l|m|})
+        Y(p, n_minus) =
+            (m & 1) ? -std::conj(Y(p, n_plus)) : std::conj(Y(p, n_plus));
+      }
+    }
+  } // end for p
+  return Y;
+}
+
+//////////////////////////////////////////////////////
+// Alternative implementations of spherical harmonics.
+// May be unstable for large l and m.
+
+/**
+ * @brief Compute real spherical harmonics \f$Y_{l m}(\theta, \phi)\f$
+ * using series expansion in cos(theta) and sin(theta). Unstable for large
+ * \f$l\f$ and \f$m\f$.
+ */
 inline double series_real_Ylm(int l, int m, double theta, double phi) {
   if (l < 0) {
     throw std::out_of_range("l must be non-negative");
@@ -696,6 +816,11 @@ inline double series_real_Ylm(int l, int m, double theta, double phi) {
                    : Xlm;
 }
 
+/**
+ * @brief Compute real spherical harmonics \f$Y_{l m}(\theta, \phi)\f$
+ * using series expansion in cos(theta) and sin(theta). Unstable for large
+ * \f$l\f$ and \f$m\f$.
+ */
 Eigen::VectorXd series_real_Ylm(int l, int m,
                                 const Eigen::MatrixXd &thetaphi_coord_P) {
 
@@ -749,6 +874,11 @@ Eigen::VectorXd series_real_Ylm(int l, int m,
   return Y;
 }
 
+/**
+ * @brief Compute spherical harmonics \f$Y_{l m}(\theta, \phi)\f$
+ * using series expansion in cos(theta) and sin(theta). Unstable for large
+ * \f$l\f$ and \f$m\f$.
+ */
 inline std::complex<double> series_Ylm(int l, int m, double theta, double phi) {
   if (l < 0) {
     throw std::out_of_range("l must be non-negative");
@@ -789,6 +919,11 @@ inline std::complex<double> series_Ylm(int l, int m, double theta, double phi) {
              : static_cast<std::complex<double>>(Xlm);
 }
 
+/**
+ * @brief Compute spherical harmonics \f$Y_{l m}(\theta, \phi)\f$
+ * using series expansion in cos(theta) and sin(theta). Unstable for large
+ * \f$l\f$ and \f$m\f$.
+ */
 Eigen::VectorXcd series_Ylm(int l, int m,
                             const Eigen::MatrixXd &thetaphi_coord_P) {
 
@@ -843,7 +978,8 @@ Eigen::VectorXcd series_Ylm(int l, int m,
 
 /**
  * @brief Compute real spherical harmonics \f$y_{l m}(\theta, \phi)\f$ for
- * \f$\ell=0,...,\ell_{max}\f$ at the given coordinates.
+ * \f$\ell=0,...,\ell_{max}\f$ at the given coordinates. Unstable for large
+ * \f$l\f$ and \f$m\f$.
  *
  * @param l_max The maximum order of the spherical harmonics (non-negative
  * integer).
@@ -935,7 +1071,8 @@ compute_all_series_real_Ylm(int l_max,
 
 /**
  * @brief Compute spherical harmonics \f$Y_{l m}(\theta, \phi)\f$ for
- * \f$\ell=0,...,\ell_{max}\f$ at the given coordinates.
+ * \f$\ell=0,...,\ell_{max}\f$ at the given coordinates. Unstable for large
+ * \f$l\f$ and \f$m\f$.
  *
  * @param l_max The maximum order of the spherical harmonics (non-negative
  * integer).
@@ -1030,6 +1167,8 @@ compute_all_series_Ylm(int l_max, const Eigen::MatrixXd &thetaphi_coord_P) {
 
 /**
  * @brief Compute Ylm(theta, phi)/exp(i*theta*phi) for m>=0 and 0<=theta<=pi.
+ * Unstable for large \f$l\f$
+ *
  * @param l Degree of the spherical harmonic.
  * @param m Order of the spherical harmonic.
  * @param theta Polar angle in radians.
@@ -1165,6 +1304,11 @@ inline double phi_independent_Ylm(int l, int m, double theta) {
   return rlm * sum_Qk;
 }
 
+/**
+ * @brief Compute spherical harmonics \f$Y_{l m}(\theta, \phi)\f$ for
+ * \f$\ell=0,...,\ell_{max}\f$ at the given coordinates. Unstable for large
+ * \f$l\f$ and \f$m\f$.
+ */
 inline std::complex<double> old_Ylm(int l, int m, double theta, double phi) {
   if (theta > M_PI || theta < 0.0) {
     throw std::out_of_range("theta must be in the range [0, pi]");
@@ -1173,6 +1317,11 @@ inline std::complex<double> old_Ylm(int l, int m, double theta, double phi) {
          phi_independent_Ylm(l, m, theta);
 }
 
+/**
+ * @brief Compute real spherical harmonics \f$y_{l m}(\theta, \phi)\f$ for
+ * \f$\ell=0,...,\ell_{max}\f$ at the given coordinates. Unstable for large
+ * \f$l\f$ and \f$m\f$.
+ */
 inline double old_real_Ylm(int l, int m, double theta, double phi) {
 
   if (theta > M_PI || theta < 0.0) {
@@ -1192,6 +1341,11 @@ inline double old_real_Ylm(int l, int m, double theta, double phi) {
   return val * std::sqrt(2) * std::cos(abs_m * phi);
 }
 
+/**
+ * @brief Compute real spherical harmonics \f$y_{l m}(\theta, \phi)\f$ for
+ * \f$\ell=0,...,\ell_{max}\f$ at the given coordinates. Unstable for large
+ * \f$l\f$ and \f$m\f$.
+ */
 Eigen::MatrixXd
 old_compute_all_real_Ylm(int l_max, const Eigen::MatrixXd &thetaphi_coord_P) {
   // check l_max
@@ -1225,8 +1379,93 @@ old_compute_all_real_Ylm(int l_max, const Eigen::MatrixXd &thetaphi_coord_P) {
 
 } // namespace special
 } // namespace mathutils
-/////////////////////////////////////
-/////////////////////////////////////
-/////////////////////////////////////
-/////////////////////////////////////
-/////////////////////////////////////
+
+namespace mathutils {} // namespace mathutils
+
+namespace mathutils {
+namespace special {
+
+/**
+ * @brief Solve for real spherical-harmonic coefficients that best fit a
+ *        3-D point cloud on a sphere, with optional Dirichlet smoothing.
+ *
+ * Solves the Tikhonov system
+ *     (Yᵀ Y + λ Lᵀ L) A = Yᵀ X,
+ * where
+ *     * Y  – design matrix of real harmonics, shape (N, (l_max+1)^2)
+ *     * L  – diagonal matrix with entries l(l+1) (Dirichlet energy weight)
+ *     * X  – input Cartesian coordinates, shape (N,3)
+ * The result A(i,:) gives the (x,y,z) coefficients for harmonic index *i*.
+ *
+ * @param XYZ0        (N,3) matrix of Cartesian samples.
+ * @param l_max       Highest harmonic degree to include.
+ * @param reg_lambda  Non-negative Tikhonov parameter (λ = 0 ⇒ plain LS).
+ * @return            ((l_max+1)^2, 3) matrix of expansion coefficients.
+ */
+inline Eigen::MatrixXd
+fit_real_sh_coefficients_to_points(const Eigen::MatrixXd &XYZ0, int l_max,
+                                   double reg_lambda = 0.0) {
+  // ---------------------------------------------------------------------
+  // 0. Basic argument checks
+  // ---------------------------------------------------------------------
+  if (l_max < 0)
+    throw std::invalid_argument(
+        "fit_real_sh_coefficients_to_points: l_max must be non-negative.");
+
+  if (XYZ0.cols() != 3)
+    throw std::invalid_argument("fit_real_sh_coefficients_to_points: XYZ0 must "
+                                "have exactly 3 columns.");
+
+  if (XYZ0.rows() == 0)
+    throw std::invalid_argument("fit_real_sh_coefficients_to_points: XYZ0 must "
+                                "contain at least one point.");
+
+  // ---------------------------------------------------------------------
+  // 1. Dimensions
+  // ---------------------------------------------------------------------
+  const int num_modes = (l_max + 1) * (l_max + 1); // (l_max+1)^2
+  const int N = static_cast<int>(XYZ0.rows());
+
+  // ---------------------------------------------------------------------
+  // 2. Build the design matrix  Y  (N × num_modes)
+  //    Only directions (θ,φ) matter; radius is ignored.
+  // ---------------------------------------------------------------------
+  Eigen::MatrixXd thetaPhi = mathutils::thetaphi_from_xyz(XYZ0); // (N,2)
+  Eigen::MatrixXd Y = compute_all_real_Ylm(l_max, thetaPhi);
+
+  // ---------------------------------------------------------------------
+  // 3. Construct Laplace–Beltrami weights  ℓ(ℓ+1)
+  // ---------------------------------------------------------------------
+  Eigen::VectorXd laplace(num_modes);
+  for (int n = 0; n < num_modes; ++n) {
+    auto [l, m] = spherical_harmonic_index_lm_N(n);
+    laplace(n) = static_cast<double>(l * (l + 1));
+  }
+
+  // ---------------------------------------------------------------------
+  // 4. Assemble the normal matrix  M = YᵀY  (SPD)  and add regularisation
+  //    directly to its diagonal:  M_ii += λ (ℓ(ℓ+1))²
+  // ---------------------------------------------------------------------
+  Eigen::MatrixXd M = Y.transpose() * Y; // (num_modes, num_modes)
+  if (reg_lambda > 0.0)
+    M.diagonal().array() += reg_lambda * laplace.array().square();
+
+  // ---------------------------------------------------------------------
+  // 5. Right-hand side  B = Yᵀ X
+  // ---------------------------------------------------------------------
+  Eigen::MatrixXd B = Y.transpose() * XYZ0; // (num_modes, 3)
+
+  // ---------------------------------------------------------------------
+  // 6. Solve  M A = B   via Cholesky (faster & stabler than explicit inverse)
+  // ---------------------------------------------------------------------
+  Eigen::LLT<Eigen::MatrixXd> chol(M);
+  if (chol.info() != Eigen::Success)
+    throw std::runtime_error("fit_real_sh_coefficients_to_points: normal "
+                             "matrix not SPD (ill-posed fit).");
+
+  Eigen::MatrixXd A = chol.solve(B); // coefficients
+  return A;                          // (num_modes, 3)
+}
+
+} // namespace special
+} // namespace mathutils
\ No newline at end of file
