@@ -4,7 +4,12 @@
 from . import jit_funs
 
 # general utility functions
-from .mathutils_backend import sign
+from .mathutils_backend import (
+    sign,
+    thetaphi_from_xyz,
+    rthetaphi_from_xyz,
+    xyz_from_rthetaphi,
+)
 
 # special functions
 from . import special
